@@ -1,24 +1,3 @@
-/*
-  Racket
-  Copyright (c) 2006-2013 PLT Design Inc.
-
-    This library is free software; you can redistribute it and/or
-    modify it under the terms of the GNU Library General Public
-    License as published by the Free Software Foundation; either
-    version 2 of the License, or (at your option) any later version.
-
-    This library is distributed in the hope that it will be useful,
-    but WITHOUT ANY WARRANTY; without even the implied warranty of
-    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
-    Library General Public License for more details.
-
-    You should have received a copy of the GNU Library General Public
-    License along with this library; if not, write to the Free
-    Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
-    Boston, MA 02110-1301 USA.
-*/
-
-
 #include "schpriv.h"
 #include "schmach.h"
 #include "future.h"
@@ -36,7 +15,7 @@ int scheme_jit_is_fixnum(Scheme_Object *rand)
       || (SAME_TYPE(SCHEME_TYPE(rand), scheme_local_type)
           && (SCHEME_GET_LOCAL_TYPE(rand) == SCHEME_LOCAL_TYPE_FIXNUM)))
     return 1;
-  else if (scheme_expr_produces_local_type(rand) == SCHEME_LOCAL_TYPE_FIXNUM)
+  else if (scheme_expr_produces_local_type(rand, NULL) == SCHEME_LOCAL_TYPE_FIXNUM)
     return 1;
   else
     return 0;
@@ -104,6 +83,7 @@ static int is_inline_unboxable_op(Scheme_Object *obj, int flag, int unsafely, in
         if (IS_NAMED_PRIM(obj, "flatan")) return 1;
         if (IS_NAMED_PRIM(obj, "fllog")) return 1;
         if (IS_NAMED_PRIM(obj, "flexp")) return 1;
+        if (IS_NAMED_PRIM(obj, "flexpt")) return 1;
       }
     }
   }
@@ -147,6 +127,7 @@ static int is_inline_unboxable_op(Scheme_Object *obj, int flag, int unsafely, in
         if (IS_NAMED_PRIM(obj, "extflatan")) return 1;
         if (IS_NAMED_PRIM(obj, "extfllog")) return 1;
         if (IS_NAMED_PRIM(obj, "extflexp")) return 1;
+        if (IS_NAMED_PRIM(obj, "extflexpt")) return 1;
       }
     }
   }
@@ -199,6 +180,7 @@ static int is_unboxing_immediate(Scheme_Object *obj, int unsafely, int extfl)
 #endif
     return unsafely;
   case scheme_toplevel_type:
+  case scheme_static_toplevel_type:
     /* Can generalize to allow any toplevel if scheme_generate_pop_unboxed() is fixed */
     if ((SCHEME_TOPLEVEL_FLAGS(obj) & SCHEME_TOPLEVEL_FLAGS_MASK) < SCHEME_TOPLEVEL_READY)
       return 0;
@@ -283,7 +265,7 @@ int scheme_can_unbox_inline(Scheme_Object *obj, int fuel, int regs, int unsafely
   }
 }
 
-int scheme_can_unbox_directly(Scheme_Object *obj, int extfl)
+int can_unbox_directly(Scheme_Object *obj, int extfl, int bfuel)
 /* Used only when !can_unbox_inline(). Detects safe operations that
    produce flonums when they don't raise an exception, and that the JIT
    supports directly unboxing. */
@@ -348,12 +330,28 @@ int scheme_can_unbox_directly(Scheme_Object *obj, int extfl)
     case scheme_letrec_type:
       obj = ((Scheme_Letrec *)obj)->body;
       break;
+    case scheme_branch_type:
+      if (!bfuel)
+        return 0;
+      bfuel--;
+      if (!can_unbox_directly(((Scheme_Branch_Rec *)obj)->tbranch, extfl, bfuel))
+        return 0;
+      obj = ((Scheme_Branch_Rec *)obj)->fbranch;
+      break;
+    case scheme_sequence_type:
+      obj = ((Scheme_Sequence *)obj)->array[((Scheme_Sequence *)obj)->count - 1];
+      break;
     default:
-      return 0;
+      return is_unboxing_immediate(obj, 0, extfl);
     }
   }
 }
 
+int scheme_can_unbox_directly(Scheme_Object *obj, int extfl)
+{
+  return can_unbox_directly(obj, extfl, 3);
+}
+
 static jit_insn *generate_arith_slow_path(mz_jit_state *jitter, Scheme_Object *rator, 
 					  jit_insn **_ref, jit_insn **_ref4,
                                           Branch_Info *for_branch, int branch_short,
@@ -665,7 +663,7 @@ static int generate_float_point_arith(mz_jit_state *jitter, Scheme_Object *rator
     } else {
 #ifdef MZ_LONG_DOUBLE
       long_double d;
-      d = long_double_from_int(second_const);
+      d = long_double_from_intptr(second_const);
       if (extfl) {
         mz_fpu_movi_ld_fppush(fpr1, d, JIT_R2)
       } else {
@@ -1116,6 +1114,7 @@ int scheme_generate_arith_for(mz_jit_state *jitter, Scheme_Object *rator, Scheme
         arith = 13 -> sqrt
         arith = 14 -> unary floating-point op (consult `rator')
         arith = 15 -> inexact->exact
+        arith = 16 -> flexpt
         cmp = 0 -> = or zero?
         cmp = +/-1 -> >=/<=
         cmp = +/-2 -> >/< or positive/negative?
@@ -1621,7 +1620,7 @@ int scheme_generate_arith_for(mz_jit_state *jitter, Scheme_Object *rator, Scheme
               __START_INNER_TINY__(branch_short);
               /* watch out for negation of most negative fixnum,
                  which is a positive number too big for a fixnum */
-              refz = jit_beqi_p(jit_forward(), JIT_R0, (void *)(((intptr_t)1 << ((8 * JIT_WORD_SIZE) - 2))));
+              refz = jit_beqi_p(jit_forward(), JIT_R0, (void *)(((uintptr_t)1 << ((8 * JIT_WORD_SIZE) - 2))));
               __END_INNER_TINY__(branch_short);
               if (reversed)
                 jit_mulr_l(JIT_R2, JIT_R0, JIT_R2);
@@ -1659,12 +1658,12 @@ int scheme_generate_arith_for(mz_jit_state *jitter, Scheme_Object *rator, Scheme
               if (!unsafe_fx || overflow_refslow) {
                 GC_CAN_IGNORE jit_insn *refx;
                 __START_INNER_TINY__(branch_short);
-                refx = jit_bnei_p(jit_forward(), JIT_R0, (void *)(((intptr_t)1 << ((8 * JIT_WORD_SIZE) - 2))));
+                refx = jit_bnei_p(jit_forward(), JIT_R0, (void *)(((uintptr_t)1 << ((8 * JIT_WORD_SIZE) - 2))));
                 __END_INNER_TINY__(branch_short);
                 /* first argument must have been most negative fixnum, 
                    second argument must have been -1: */
                 if (reversed)
-                  (void)jit_movi_p(JIT_R0, (void *)(((intptr_t)1 << ((8 * JIT_WORD_SIZE) - 1)) | 0x1));
+                  (void)jit_movi_p(JIT_R0, (void *)(((uintptr_t)1 << ((8 * JIT_WORD_SIZE) - 1)) | 0x1));
                 else
                   (void)jit_movi_p(JIT_R0, scheme_make_integer(-1));
                 (void)jit_jmpi(refslow);
@@ -1781,10 +1780,10 @@ int scheme_generate_arith_for(mz_jit_state *jitter, Scheme_Object *rator, Scheme
           /* Non-constant arg is in JIT_R0 */
           if (arith == ARITH_ADD) {
             if (unsafe_fx && !overflow_refslow)
-              jit_addi_l(dest, JIT_R0, v << 1);
+              jit_addi_l(dest, JIT_R0, (uintptr_t)v << 1);
             else {
               jit_movr_p(JIT_R2, JIT_R0);
-              (void)jit_boaddi_l(refslow, JIT_R2, v << 1);
+              (void)jit_boaddi_l(refslow, JIT_R2, (uintptr_t)v << 1);
               jit_movr_p(dest, JIT_R2);
             }
           } else if (arith == ARITH_SUB) {
@@ -1797,10 +1796,10 @@ int scheme_generate_arith_for(mz_jit_state *jitter, Scheme_Object *rator, Scheme
               jit_addi_ul(dest, JIT_R2, 0x1);
             } else {
               if (unsafe_fx && !overflow_refslow)
-                jit_subi_l(dest, JIT_R0, v << 1);
+                jit_subi_l(dest, JIT_R0, (uintptr_t)v << 1);
               else {
                 jit_movr_p(JIT_R2, JIT_R0);
-                (void)jit_bosubi_l(refslow, JIT_R2, v << 1);
+                (void)jit_bosubi_l(refslow, JIT_R2, (uintptr_t)v << 1);
                 jit_movr_p(dest, JIT_R2);
               }
             }
@@ -1832,7 +1831,7 @@ int scheme_generate_arith_for(mz_jit_state *jitter, Scheme_Object *rator, Scheme
               jit_ori_ul(dest, JIT_R0, l);
             } else if (arith == ARITH_XOR) {
               /* xor */
-              jit_xori_ul(dest, JIT_R0, v << 1);
+              jit_xori_ul(dest, JIT_R0, (uintptr_t)v << 1);
             } else if ((arith == ARITH_LSH) || (arith == ARITH_RSH)) {
               /* arithmetic-shift */
               /* We only get here when v is between -MAX_TRY_SHIFT and MAX_TRY_SHIFT, inclusive */
@@ -1883,7 +1882,7 @@ int scheme_generate_arith_for(mz_jit_state *jitter, Scheme_Object *rator, Scheme
               __END_INNER_TINY__(branch_short);
               /* watch out for most negative fixnum! */
               if (!unsafe_fx || overflow_refslow)
-                (void)jit_beqi_p(refslow, JIT_R0, (void *)(((intptr_t)1 << ((8 * JIT_WORD_SIZE) - 1)) | 0x1));
+                (void)jit_beqi_p(refslow, JIT_R0, (void *)(((uintptr_t)1 << ((8 * JIT_WORD_SIZE) - 1)) | 0x1));
               (void)jit_movi_p(JIT_R1, scheme_make_integer(0));
               jit_subr_l(JIT_R0, JIT_R1, JIT_R0);
               jit_ori_l(JIT_R0, JIT_R0, 0x1);
@@ -2021,7 +2020,7 @@ int scheme_generate_arith_for(mz_jit_state *jitter, Scheme_Object *rator, Scheme
             jit_lshr_l(JIT_R0, JIT_V1, JIT_R0);
             ref3 = jit_bmcr_l(jit_forward(), JIT_R1, JIT_R0);
           } else {
-            ref3 = jit_bmci_l(jit_forward(), JIT_R0, 1 << (v+1));
+            ref3 = jit_bmci_l(jit_forward(), JIT_R0, (uintptr_t)1 << (v+1));
             rs_can_keep = 1;
           }
           break;
@@ -2108,7 +2107,7 @@ int scheme_generate_extflonum_arith(mz_jit_state *jitter, Scheme_Object *rator,
 }
 
 
-#define MAX_NON_SIMPLE_ARGS 5
+#define MAX_NON_SIMPLE_ARGS 6
 
 static int extract_nary_arg(int reg, int n, mz_jit_state *jitter, Scheme_App_Rec *app, 
                             Scheme_Object **alt_args, int old_short_jumps)
@@ -2119,7 +2118,7 @@ static int extract_nary_arg(int reg, int n, mz_jit_state *jitter, Scheme_App_Rec
       scheme_generate_unboxing(jitter, JIT_R0);
   } else if (scheme_is_constant_and_avoids_r1(app->args[n+1])) {
     __END_SHORT_JUMPS__(old_short_jumps);
-    scheme_generate(app->args[n+1], jitter, 0, 0, 0, reg, NULL);
+    scheme_generate(app->args[n+1], jitter, 0, 0, 0, reg, NULL, NULL);
     CHECK_LIMIT();
     __START_SHORT_JUMPS__(old_short_jumps);
   } else {
@@ -2165,18 +2164,19 @@ static void patch_nary_branches(mz_jit_state *jitter, Branch_Info *for_nary_bran
 
 int scheme_generate_nary_arith(mz_jit_state *jitter, Scheme_App_Rec *app,
                                int arith, int cmp, Branch_Info *for_branch, int branch_short,
+                               int unsafe_fx, int unsafe_fl,
                                int dest)
 {
-  int c, i, non_simple_c = 0, stack_c, use_fx = 1, trigger_arg = 0;
-  Scheme_Object *non_simples[1+MAX_NON_SIMPLE_ARGS], **alt_args, *v;
+  int c, i, non_simple_c = 0, stack_c, use_fx = !unsafe_fl, trigger_arg = 0, use_short;
+  Scheme_Object *non_simples[MAX_NON_SIMPLE_ARGS], **alt_args, *v;
   Branch_Info for_nary_branch;
   Branch_Info_Addr nary_addrs[3];
   GC_CAN_IGNORE jit_insn *refslow, *reffx, *refdone;
   GC_CAN_IGNORE jit_insn *reffalse = NULL, *refdone3 = NULL;
-#ifdef INLINE_FP_OPS
+#ifdef INLINE_FP_COMP
   int args_unboxed;
   GC_CAN_IGNORE jit_insn *reffl, *refdone2;
-  int use_fl = 1;
+  int use_fl = !unsafe_fx;
 # define mzSET_USE_FL(x) x
 #else
 # define mzSET_USE_FL(x) /* empty */
@@ -2189,11 +2189,43 @@ int scheme_generate_nary_arith(mz_jit_state *jitter, Scheme_App_Rec *app,
     mzSET_USE_FL(use_fl = 0);
   }
 
+#ifdef INLINE_FP_COMP
+# ifndef INLINE_FP_OPS
+  if ((arith == ARITH_ADD)
+      || (arith == ARITH_SUB)
+      || (arith == ARITH_MUL)
+      || (arith == ARITH_DIV)) {
+    /* assert: unsafe_fl < 1 */
+    use_fl = 0;
+  }
+# endif
+#endif
+
   c = app->num_args;
+  if (!c) {
+    /* Constant folding would normally prevent us from getting here, but just in case */
+    if ((arith == ARITH_ADD) || (arith == ARITH_IOR) || (arith == ARITH_XOR)) {
+      if (!unsafe_fl)
+        (void)jit_movi_p(dest, scheme_make_integer(0));
+      else
+        (void)jit_movi_p(dest, scheme_zerod);
+      return 1;
+    } else if (arith == ARITH_AND) {
+      (void)jit_movi_p(dest, scheme_make_integer(-1));
+      return 1;
+    } else if (arith == ARITH_MUL) {
+      if (!unsafe_fl)
+        (void)jit_movi_p(dest, scheme_make_integer(1));
+      else
+        scheme_mz_load_retained(jitter, dest, scheme_make_double(1.0));
+      return 1;
+    }
+  }
+
   for (i = 0; i < c; i++) {
     v = app->args[i+1];
     if (!scheme_is_constant_and_avoids_r1(v)) {
-      if (non_simple_c < MAX_NON_SIMPLE_ARGS)
+      if (non_simple_c < (MAX_NON_SIMPLE_ARGS-1))
         non_simples[1+non_simple_c] = v;
       non_simple_c++;
     }
@@ -2205,13 +2237,13 @@ int scheme_generate_nary_arith(mz_jit_state *jitter, Scheme_App_Rec *app,
       use_fx = 0;
       if (trigger_arg == i)
         trigger_arg++;
-    } else if (SCHEME_TYPE(v) >= _scheme_compiled_values_types_) {
+    } else if (SCHEME_TYPE(v) >= _scheme_ir_values_types_) {
       use_fx = 0;
       mzSET_USE_FL(use_fl = 0);
     }
   }
 
-  if ((non_simple_c <= MAX_NON_SIMPLE_ARGS) && (non_simple_c < c)) {
+  if ((non_simple_c <= (MAX_NON_SIMPLE_ARGS-1)) && (non_simple_c < c)) {
     stack_c = non_simple_c;
     alt_args = non_simples;
     non_simples[0] = app->args[0];
@@ -2222,11 +2254,12 @@ int scheme_generate_nary_arith(mz_jit_state *jitter, Scheme_App_Rec *app,
   }
 
   if (stack_c)
-    scheme_generate_app(app, alt_args, stack_c, jitter, 0, 0, 0, 2);
+    scheme_generate_app(app, alt_args, stack_c, stack_c, jitter, 0, 0, 0, 2);
   CHECK_LIMIT();
   mz_rs_sync();
 
-  __START_SHORT_JUMPS__(c < 100);
+  use_short = c < 100;
+  __START_SHORT_JUMPS__(use_short);
 
   if (trigger_arg >= c) {
     /* we don't expect this to happen, since constant-folding normally
@@ -2235,13 +2268,17 @@ int scheme_generate_nary_arith(mz_jit_state *jitter, Scheme_App_Rec *app,
     trigger_arg = 0;
   }
 
-  extract_nary_arg(JIT_R0, trigger_arg, jitter, app, alt_args, c < 100);
+  extract_nary_arg(JIT_R0, trigger_arg, jitter, app, alt_args, use_short);
   CHECK_LIMIT();
-  /* trigger argument a fixnum? */
-  reffx = jit_bmsi_ul(jit_forward(), JIT_R0, 0x1);
 
-#ifdef INLINE_FP_OPS
-  if (use_fl) {
+  if (unsafe_fl < 1) {
+    /* trigger argument a fixnum? */
+    reffx = jit_bmsi_ul(jit_forward(), JIT_R0, 0x1);
+  } else
+    reffx = NULL;
+
+#ifdef INLINE_FP_COMP
+  if (use_fl && (unsafe_fl < 1)) {
     /* First argument a flonum? */
     jit_ldxi_s(JIT_R0, JIT_R0, &((Scheme_Object *)0x0)->type);
     reffl = jit_beqi_i(jit_forward(), JIT_R0, scheme_double_type);
@@ -2251,53 +2288,76 @@ int scheme_generate_nary_arith(mz_jit_state *jitter, Scheme_App_Rec *app,
   }
 #endif
   
-  if (!use_fx) {
+  if (!use_fx && reffx) {
     mz_patch_branch(reffx);
   }
 
-  refslow = jit_get_ip();
-  /* slow path */
-  if (alt_args) {
-    /* get all args on runstack */
-    int delta = stack_c - c;
-    for (i = 0; i < c; i++) {
-      if (delta) {
-        extract_nary_arg(JIT_R0, i, jitter, app, alt_args, c < 100);
-        CHECK_LIMIT();
-        jit_stxi_p(WORDS_TO_BYTES(i+delta), JIT_RUNSTACK, JIT_R0);
-      } else
-        break;
+  if ((unsafe_fl < 1) && (unsafe_fx < 1)) {
+    refslow = jit_get_ip();
+    /* slow path */
+    if (alt_args) {
+      /* get all args on runstack */
+      int delta = stack_c - c;
+      for (i = 0; i < c; i++) {
+        if (delta) {
+          extract_nary_arg(JIT_R0, i, jitter, app, alt_args, use_short);
+          CHECK_LIMIT();
+          jit_stxi_p(WORDS_TO_BYTES(i+delta), JIT_RUNSTACK, JIT_R0);
+        } else
+          break;
+      }
+      jit_subi_p(JIT_RUNSTACK, JIT_RUNSTACK, WORDS_TO_BYTES(c - stack_c));
     }
-    jit_subi_p(JIT_RUNSTACK, JIT_RUNSTACK, WORDS_TO_BYTES(c - stack_c));
+    (void)jit_movi_p(JIT_V1, ((Scheme_Primitive_Proc *)app->args[0])->prim_val);
+    (void)jit_movi_i(JIT_R1, c);
+    (void)jit_calli(sjc.call_original_nary_arith_code);
+    if (alt_args) {
+      jit_addi_p(JIT_RUNSTACK, JIT_RUNSTACK, WORDS_TO_BYTES(c - stack_c));
+    }
+    refdone = jit_jmpi(jit_forward());
+  } else {
+    refdone = NULL;
+    refslow = NULL;
   }
-  (void)jit_movi_p(JIT_V1, ((Scheme_Primitive_Proc *)app->args[0])->prim_val);
-  (void)jit_movi_i(JIT_R1, c);
-  (void)jit_calli(sjc.call_original_nary_arith_code);
-  if (alt_args) {
-    jit_addi_p(JIT_RUNSTACK, JIT_RUNSTACK, WORDS_TO_BYTES(c - stack_c));
-  }
-  refdone = jit_jmpi(jit_forward());
+
   if (!arith) {
+    GC_CAN_IGNORE jit_insn *refskip;
+    if ((unsafe_fx > 0) || (unsafe_fl > 0)) {
+      /* No dispatch so far, so jump to fast path to skip #f result */
+      __START_INNER_TINY__(use_short);
+      refskip = jit_jmpi(jit_forward());
+      __END_INNER_TINY__(use_short);
+    } else
+      refskip = NULL;
+    
     reffalse = jit_get_ip();
     (void)jit_movi_p(JIT_R0, scheme_false);
     refdone3 = jit_jmpi(jit_forward());
+
+    if (refskip) {
+      __START_INNER_TINY__(use_short);
+      mz_patch_ucbranch(refskip);
+      __END_INNER_TINY__(use_short);
+    }
   } else {
     reffalse = NULL;
   }
 
-#ifdef INLINE_FP_OPS
+#ifdef INLINE_FP_COMP
   if (use_fl) {
     /* Flonum branch: */
-    mz_patch_branch(reffl);
-    for (i = 0; i < c; i++) {
-      if (i != trigger_arg) {
-        v = app->args[i+1];
-        if (!SCHEME_FLOATP(v)) {
-          extract_nary_arg(JIT_R0, i, jitter, app, alt_args, c < 100);
-          (void)jit_bmsi_ul(refslow, JIT_R0, 0x1);
-          jit_ldxi_s(JIT_R0, JIT_R0, &((Scheme_Object *)0x0)->type);
-          (void)jit_bnei_i(refslow, JIT_R0, scheme_double_type);
-          CHECK_LIMIT();
+    if (unsafe_fl < 1) {
+      mz_patch_branch(reffl);
+      for (i = 0; i < c; i++) {
+        if (i != trigger_arg) {
+          v = app->args[i+1];
+          if (!SCHEME_FLOATP(v)) {
+            extract_nary_arg(JIT_R0, i, jitter, app, alt_args, use_short);
+            (void)jit_bmsi_ul(refslow, JIT_R0, 0x1);
+            jit_ldxi_s(JIT_R0, JIT_R0, &((Scheme_Object *)0x0)->type);
+            (void)jit_bnei_i(refslow, JIT_R0, scheme_double_type);
+            CHECK_LIMIT();
+          }
         }
       }
     }
@@ -2305,19 +2365,19 @@ int scheme_generate_nary_arith(mz_jit_state *jitter, Scheme_App_Rec *app,
     args_unboxed = ((arith != ARITH_MIN) && (arith != ARITH_MAX)); /* no unboxing for min & max */
     if (args_unboxed)
       jitter->unbox++;
-    extract_nary_arg(JIT_R0, 0, jitter, app, alt_args, c < 100);
+    extract_nary_arg(JIT_R0, 0, jitter, app, alt_args, use_short);
     CHECK_LIMIT();
     for (i = 1; i < c; i++) {
       if (!arith && (i > 1))
-        extract_nary_arg(JIT_R0, i - 1, jitter, app, alt_args, c < 100);
-      extract_nary_arg((args_unboxed ? JIT_R0 : JIT_R1), i, jitter, app, alt_args, c < 100);
+        extract_nary_arg(JIT_R0, i - 1, jitter, app, alt_args, use_short);
+      extract_nary_arg((args_unboxed ? JIT_R0 : JIT_R1), i, jitter, app, alt_args, use_short);
       if ((i == c - 1) && args_unboxed) --jitter->unbox; /* box last result */
       if (!arith) init_nary_branches(&for_nary_branch, nary_addrs);
-      __END_SHORT_JUMPS__(c < 100);
+      __END_SHORT_JUMPS__(use_short);
       scheme_generate_arith(jitter, NULL, NULL, scheme_void, 2, arith, cmp, 0,
-                            !arith ? &for_nary_branch : NULL, c < 100, 0, 1, NULL,
+                            !arith ? &for_nary_branch : NULL, use_short, 0, 1, NULL,
                             JIT_R0);
-      __START_SHORT_JUMPS__(c < 100);
+      __START_SHORT_JUMPS__(use_short);
       if (!arith) patch_nary_branches(jitter, &for_nary_branch, reffalse);
       CHECK_LIMIT();
     }
@@ -2333,38 +2393,40 @@ int scheme_generate_nary_arith(mz_jit_state *jitter, Scheme_App_Rec *app,
 
   if (use_fx) {
     /* Fixnum branch */
-    mz_patch_branch(reffx);
-    for (i = 0; i < c; i++) {
-      if (i != trigger_arg) {
-        v = app->args[i+1];
-        if (!SCHEME_INTP(v)) {
-          extract_nary_arg(JIT_R0, i, jitter, app, alt_args, c < 100);
-          CHECK_LIMIT();
-          (void)jit_bmci_ul(refslow, JIT_R0, 0x1);
-          CHECK_LIMIT();
+    if (unsafe_fx < 1) {
+      mz_patch_branch(reffx);
+      for (i = 0; i < c; i++) {
+        if (i != trigger_arg) {
+          v = app->args[i+1];
+          if (!SCHEME_INTP(v)) {
+            extract_nary_arg(JIT_R0, i, jitter, app, alt_args, use_short);
+            CHECK_LIMIT();
+            (void)jit_bmci_ul(refslow, JIT_R0, 0x1);
+            CHECK_LIMIT();
+          }
         }
       }
     }
     /* All fixnums, so inline fast fixnum combination;
        on overflow, bail out to refslow. */
-    extract_nary_arg(JIT_R0, 0, jitter, app, alt_args, c < 100);
+    extract_nary_arg(JIT_R0, 0, jitter, app, alt_args, use_short);
     for (i = 1; i < c; i++) {
       if (!arith && (i > 1))
-        extract_nary_arg(JIT_R0, i - 1, jitter, app, alt_args, c < 100);
-      extract_nary_arg(JIT_R1, i, jitter, app, alt_args, c < 100);
+        extract_nary_arg(JIT_R0, i - 1, jitter, app, alt_args, use_short);
+      extract_nary_arg(JIT_R1, i, jitter, app, alt_args, use_short);
       CHECK_LIMIT();
       if (!arith) init_nary_branches(&for_nary_branch, nary_addrs);
-      __END_SHORT_JUMPS__(c < 100);
+      __END_SHORT_JUMPS__(use_short);
       scheme_generate_arith(jitter, NULL, NULL, scheme_void, 2, arith, cmp, 0,
-                            !arith ? &for_nary_branch : NULL, c < 100, 1, 0, refslow,
+                            !arith ? &for_nary_branch : NULL, use_short, 1, 0, refslow,
                             JIT_R0);
-      __START_SHORT_JUMPS__(c < 100);
+      __START_SHORT_JUMPS__(use_short);
       if (!arith) patch_nary_branches(jitter, &for_nary_branch, reffalse);
       CHECK_LIMIT();
     }
   }
 
-#ifdef INLINE_FP_OPS
+#ifdef INLINE_FP_COMP
   if (use_fl && use_fx) {
     mz_patch_ucbranch(refdone2);
   }
@@ -2372,11 +2434,12 @@ int scheme_generate_nary_arith(mz_jit_state *jitter, Scheme_App_Rec *app,
   if (!arith) {
     (void)jit_movi_p(JIT_R0, scheme_true);
   }
-  mz_patch_ucbranch(refdone);
+  if (refdone)
+    mz_patch_ucbranch(refdone);
   if (refdone3)
     mz_patch_ucbranch(refdone3);
 
-  __END_SHORT_JUMPS__(c < 100);
+  __END_SHORT_JUMPS__(use_short);
 
   if (stack_c) {
     mz_rs_inc(stack_c); /* no sync */
