@@ -1,4 +1,4 @@
-/* 
+/*
    The "cstartup.inc" file is not, as distributed, compatible with
    omissions from or additions to the set of built-in identifiers.
    The reason is that the "cstartup.inc" file is a ".zo" version of the
@@ -14,24 +14,9 @@
 
 #define USE_COMPILED_STARTUP 1
 
-#define EXPECTED_PRIM_COUNT 1125
-#define EXPECTED_UNSAFE_COUNT 106
-#define EXPECTED_FLFXNUM_COUNT 69
-#define EXPECTED_EXTFL_COUNT 45
-#define EXPECTED_FUTURES_COUNT 15
-#define EXPECTED_FOREIGN_COUNT 79
+#define EXPECTED_PRIM_COUNT 1452
 
 #ifdef MZSCHEME_SOMETHING_OMITTED
 # undef USE_COMPILED_STARTUP
 # define USE_COMPILED_STARTUP 0
 #endif
-
-#if defined(__MWERKS__) && !defined(powerc)
-#define MZCOMPILED_STRING_FAR far
-#else
-#define MZCOMPILED_STRING_FAR /**/
-#endif
-
-#if USE_COMPILED_STARTUP
-extern Scheme_Object *scheme_eval_compiled_sized_string(const char *str, int len, Scheme_Env *env);
-#endif
